import requests
from sequence import Sequence
import regex
from errors import NoResultsError, InvalidQueryError, TooManyResultsError
from partialmethod import partialmethod


class OEISClient(object):
    '''Maintains a :class:`Session<requests:requests.Session>` and
       contains all methods for querying the OEIS.'''

    SEARCH_URL = 'http://oeis.org/search'

    def __init__(self):
        self.session = requests.Session()
        self.session.params = {'fmt': 'text'}

    def _check_response(self, response_text, query):
        '''Checks reponses from the OEIS, raises relevant exception
           upon error.'''

        # if no results are found, raise appropriate exception
        if response_text.find('\nNo results.\n') != -1:
            raise NoResultsError(query)

        # if query has invalid syntax, raise appropriate exception
        if response_text.find('Could not parse search query:') != -1:
            raise InvalidQueryError(response_text)

        # if there are too many results to display, raise appropriate exception
        if response_text.find('Too many results.') != -1:
            raise TooManyResultsError(query)

    def _parse_response(self, response_text):
        '''Takes a multi-sequence internal format response & builds
           Sequence objects for each sequence. Returns a list of the
           Sequence objects.'''

        blankline_split = regex.blank_line.split(response_text)
        seqs = []
        if blankline_split[1][1] == u'No results.':
            return seqs

        for entry in blankline_split[2:-1]:      # stripping header & footer
            seqs.append(Sequence(entry))

        return seqs

    def lookup_by(self, prefix, query, max_seqs=10, list_func=False):
        '''If *prefix* is `""`, search OEIS with string *query*,
           otherwise use string '*prefix*:*query*'.

           If *list_func* is true, return a list of at most
           *max_seqs* :class:`Sequence <sequence.Sequence>` objects or
           else an empty list if there are no results. If *list_func*
           is false, return the first Sequence found, or else raise a
           :exc:`NoResultsError <errors.NoResultsError>`.'''

        if not prefix:
            search_string = query
        else:
            search_string = '{}:{}'.format(prefix, query)

        try:
            response = self.session.get(self.SEARCH_URL,
                                        params={'q': search_string})
            self._check_response(response.text, search_string)
            seqs = self._parse_response(response.text)
            num_seqs = regex.showing_line.search(response.text).groups()[0]
            for i in xrange(10, (int(num_seqs) - (int(num_seqs) % 10)) + 1):
                if len(seqs) < max_seqs:
                    response = self.session.get(self.SEARCH_URL,
                                                params={'q': search_string,
                                                        'start': str(i)})
                    self._check_response(response.text, search_string)
                    seqs.extend(self._parse_response(response.text))
                else:
                    break

            if len(seqs) > max_seqs:
                seqs = seqs[:max_seqs]

            if list_func:
                return seqs
            else:
                return seqs[0]

        except NoResultsError:
            if list_func:
                return []
            else:
                raise NoResultsError(search_string)

    get_by_id = partialmethod(lookup_by, 'id')
    get_by_id.__doc__ = '''Returns a
                           :class:`Sequence <sequence.Sequence>` for
                           the sequence of the specified ID, else
                           raises :exc:`NoResultsError`.'''
    lookup_by_name = partialmethod(lookup_by, 'name', list_func=True)
    lookup_by_name.__doc__ = '''Returns a list of
                                :class:`Sequence <sequence.Sequence>`
                                objects whose names contain *name*.'''
    lookup_by_author = partialmethod(lookup_by, 'author', list_func=True)
    lookup_by_author.__doc__ = '''Returns a list of
                                  :class:`Sequence <sequence.Sequence>`
                                  objects whose authors contain
                                  *author*.'''

    def lookup_by_keywords(self, keywords):
        '''Returns a list of at most *max_seqs*
           :class:`Sequence <sequence.Sequence>` objects which are
           tagged with *keywords*.'''

        query = '"'+' '.join(keywords)+'"'
        return self.lookup_by('keyword', query, list_func=True)

    def lookup_by_terms(self, terms, **kwargs):
        '''Returns a list of at most *max_seqs*
           :class:`Sequence <sequence.Sequence>` objects which
           contain *terms* anywhere within them. If none exist,
           returns an empty list. If *ordered* is false, terms may be
           in any order. If *signed* is false, terms may be positive
           or negative.'''

        # if order does not matter, specify in search query (space-delimited)
        if ('ordered' in kwargs) and not kwargs['ordered']:
            query = ' '.join(terms)
        else:
            # ordered is assumed True
            query = ','.join(map(str, terms))

        # if sign does not matter, specify in search query (space-delimited)
        if ('signed' in kwargs) and not kwargs['signed']:
            return self.lookup_by('seq', query, list_func=True)
        else:
            # signed is assumed True
            return self.lookup_by('signed', query, list_func=True)

    def extend_sequence(self, *terms):
        '''Returns a :class:`Sequence <sequence.Sequence>` object for
           the first sequence, sorting by relevance, which contains
           *terms* consecutively.'''

        seqs = self.lookup_by_terms(*terms)
        return seqs[0]



class OEISException(Exception):
    '''Base class for PyOEIS exceptions.'''
    pass


class NoResultsError(OEISException):
    '''Raised when a search gives no results and it is unacceptable to
       return an empty list.'''
    def __init__(self, query):
        self.query = query

    def __str__(self):
        return 'No results for query "'+self.query+'"'


class TooManyResultsError(OEISException):
    '''Raised when too many results are found for a search for them to
       be properly parsed.'''
    def __init__(self, query):
        self.query = query

    def __str__(self):
        ret = ('Too many results for query "' + self.query
               + '". Narrow your search.')
        return ret


class InvalidQueryError(OEISException):
    '''Raised when a search is invalid according to the OEIS search
       syntax.'''
    def __init__(self, response):
        self.response = response[response.find(':')+2:]

    def __str__(self):
        return self.response


class MalformedSequenceError(OEISException):
    '''Raised when a sequence entry does not contain the required
       information in a parsable format.'''
    def __init__(self, malformed_line):
        self.malformed_line = malformed_line

    def __str__(self):
        return ("Line '" + self.malformed_line +
                "' is malformed in the sequence entry.")


class NoFunctionError(OEISException):
    '''Raised when a sequence has no formula.'''
    def __init__(self, seq):
        self.seq_name = seq.name

    def __str__(self):
        return "The sequence '" + self.seq_name + "' has no formula listed."



import re

# OEIS Internal Format parsing
id = re.compile(ur'^%I (A\d+) ?([MN]\d+)? ?([MN]\d+)?', re.M)
unsigned = re.compile(ur'^%[STU] A\d+ ((?:\d+,?)*)', re.M)
signed = re.compile(ur'^%[VWX] A\d+ ((?:-?\d+,?)*)', re.M)
name = re.compile(ur'^%N A\d+ ([^.:]+)([^.]*)', re.M)
reference = re.compile(ur'^%D A\d+ (.+).?', re.M)
link = re.compile(ur'^%H A\d+ (.*)<a href="(.*)">(.*)</a>(.*)', re.M)
formula = re.compile(ur'^%F A\d+ (.+)', re.M)
cross_reference = re.compile(ur'^%Y A\d+ (.+)\.?', re.M)
author = re.compile(ur'^%A A\d+ (.+)', re.M)
offset = re.compile(ur'^%O A\d+ (-?\d+), ?(-?\d+)', re.M)
maple = re.compile(ur'^%p A\d+ (.+)', re.M)
mathematica = re.compile(ur'^%t A\d+ (.+)', re.M)
other_programs = re.compile(ur'^%o A\d+ (.+)', re.M)
error = re.compile(ur'^%E A\d+ (.+)\.?', re.M)
example = re.compile(ur'^%e A\d+ (.+)', re.M)
keywords = re.compile(ur'%K A\d+ [a-zA-Z]+,?', re.M)
comment = re.compile(ur'%C A\d+ (.+)', re.M)
blank_line = re.compile(ur'^\n\s*', re.M)
showing_line = re.compile(ur'Showing \d+-\d+ of (\d+)')

# Maths expression parsing
valid = re.compile(ur'^[()^+-/*!n\d]+$')
parens = re.compile(ur'\([+-/n\d*!^]+\)')
operands = re.compile(ur'(\d+|n)')
fac = re.compile(ur'(\(?[+\d*-/n^]+\)?)!')
opp_paren = re.compile(ur'\)\(')
paren_fac = re.compile(ur'\)fac')
expo = re.compile(ur'\^')



from functools import partial


class partialmethod(partial):
    def __get__(self, instance, owner):
        if instance is None:
            return self
        return partial(self.func, instance,
                       *(self.args or ()), **(self.keywords or {}))



import regex
from helpers import parse_comma_separated_findall
from errors import MalformedSequenceError, NoFunctionError
from math import factorial as fac  # NOQA


class Sequence(object):
    '''Takes internal format sequence entry as constructor argument.

       Has attributes to contain information for each field of a
       sequence entry in the OEIS and methods fot retrieving a certain
       number of the sequence's signed or unsigned terms.'''

    def __init__(self, sequence_entry):
        # parse %I
        try:
            id_search = regex.id.search(sequence_entry)
            self.id = id_search.groups()[0]
        except AttributeError:
            raise MalformedSequenceError('%I')
        self.alt_ids = []
        if id_search.groups()[1] is not None:
            self.alt_ids.extend(id_search.groups()[1:])

        # parse %S, %T and %U
        try:
            unsigned_findall = regex.unsigned.findall(sequence_entry)
            unsigned_strs = parse_comma_separated_findall(unsigned_findall)
            self.unsigned_list = map(int, unsigned_strs)
        except AttributeError:
            raise MalformedSequenceError('%S')

        # parse %V, %W and %X
        try:
            signed_findall = regex.signed.findall(sequence_entry)
            signed_strs = parse_comma_separated_findall(signed_findall)
            self.signed_list = map(int, signed_strs)
        except AttributeError:     # sequence entry has no %V, %W or %X
            self.signed_list = []

        # parse %N
        try:
            name_search = regex.name.search(sequence_entry)
            self.name = name_search.groups()[0]
            self.raw_formula = name_search.groups()[1]
        except AttributeError:
            raise MalformedSequenceError('%N')

        # parse %D
        try:
            reference_findall = regex.reference.findall(sequence_entry)
            self.references = reference_findall
        except AttributeError:     # sequence entry has no %D
            self.references = []

        # parse %H
        try:
            link_findall = regex.link.findall(sequence_entry)
            for l in link_findall:
                url = l[1]
                text = l[0] + l[2] + l[3]
                self.links.append({'text': text, 'url': url})
        except AttributeError:     # sequence entry has no %H
            self.links = []

        # parse %F
        try:
            formula_findall = regex.formula.findall(sequence_entry)
            self.formulae = formula_findall
        except AttributeError:     # sequence entry has no %F
            self.formulae = []

        # parse %Y
        try:
            crossref_findall = regex.cross_reference.findall(sequence_entry)
            self.cross_references = crossref_findall
        except AttributeError:     # sequence entry has no %Y
            self.cross_references = []

        # parse %A
        try:
            author_search = regex.author.search(sequence_entry)
            self.author = author_search.groups()[0]
        except AttributeError:
            raise MalformedSequenceError('%A')

        # parse %O
        try:
            offset_search = regex.offset.search(sequence_entry)
            self.offset = tuple(map(int, offset_search.groups()))
        except AttributeError:
            raise MalformedSequenceError('%O')

        # parse %E
        try:
            error_findall = regex.error.findall(sequence_entry)
            self.errors = error_findall
        except AttributeError:     # sequence entry has no %E
            self.errors = []

        # parse %e
        try:
            example_findall = regex.example.findall(sequence_entry)
            self.examples = example_findall
        except AttributeError:     # sequence entry has no %e
            self.examples = []

        # parse %p
        try:
            maple_search = regex.maple.search(sequence_entry)
            self.maple = maple_search.groups()[0]
        except AttributeError:     # sequence entry has no %p
            self.maple = None

        # parse %t
        try:
            mathematica_search = regex.mathematica.search(sequence_entry)
            self.mathematica = mathematica_search.groups()[0]
        except AttributeError:     # sequence entry has no %t
            self.mathematica = None

        # parse %o
        try:
            otherprogs_findall = regex.other_programs.findall(sequence_entry)
            self.other_programs = otherprogs_findall
        except AttributeError:     # sequence entry has no %o
            self.other_programs = []

        # parse %K
        try:
            keywords_findall = regex.keywords.findall(sequence_entry)
            self.keywords = parse_comma_separated_findall(keywords_findall)
        except AttributeError:
            raise MalformedSequenceError('%K')

        # parse %C
        try:
            comment_findall = regex.comment.findall(sequence_entry)
            self.comments = comment_findall
        except AttributeError:     # sequence entry has no %C
            self.comments = []

        # create generate method
        self.generate = self._make_generate()
        self.generate.__doc__ = """
                                If a parsable formula exists, returns
                                the *n*th term of the sequence, else
                                raises a
                                :exc:`NoFunctionError <errors.NoFunctionError>`
                                """

    def __str__(self):
        return '<Sequence object for "' + self.name + '">'

    def _make_generate(self):
        raw = self.raw_formula
        raw = raw.replace(u' ', u'')
        candidates = raw.split(u'=')
        valid_formula = None
        for s in candidates:
            valid = regex.valid.match(s) is not None
            if valid:
                valid_formula = s
                break

        if valid_formula is None:
            def generate(n):
                raise NoFunctionError(self)
        else:
            def make_evaluable(expr):
                parens = (p[1:-1] for p in regex.parens.findall(expr))
                evaluable = expr
                for pair in zip(parens, map(make_evaluable, parens)):
                    evaluable = evaluable.replace(pair[0], pair[1])
                # Handle factorials ('!')
                evaluable = regex.fac.sub(lambda m: '(fac('+m.groups()[0]+'))',
                                          evaluable)
                # Parenthesise operands (integers and 'n')
                evaluable = regex.operands.sub(lambda m: '('+m.groups()[0]+')',
                                               evaluable)
                # Handle implictit ('(2)(n)' and 'fac(i)fac(j)')
                evaluable = regex.opp_paren.sub(')*(', evaluable)
                # Handle exponentiation ('^')
                evaluable = regex.expo.sub('**', evaluable)
                print 'foo'
                return evaluable

            to_eval = make_evaluable(valid_formula)

            def generate(n):
                return eval(to_eval)

        return generate

    def unsigned(self, n):
        '''Returns the first *n* unsigned integers in the sequence.'''
        return self.unsigned_list[:n]

    def signed(self, n):
        '''Returns the first *n* signed integers in the sequence.'''
        return self.signed_list[:n]



from distutils.core import setup
import sys


extra = {}
if sys.version_info >= (3, ):
    extra['use_2to3'] = True

setup(
    name='pyoeis',
    packages=['pyoeis', ],
    version='1.1.1',
    author='Dylan Evans',
    author_email='dylan@physicynicism.com',
    url='www.physicynicism.com',
    description='Python interface to the OEIS',
    long_description='''
    PyOEIS provides several methods to search the `Online
    Encyclopedia of Integer Sequences with Python (OEIS)
    <http://www.oeis.org>`_. It also allows for access
    the information contained within individual sequence
    entries, obtained by the querying methods
    provided.''',
    classifiers=['Development Status :: 3 - Alpha',
                 'License :: OSI Approved :: Apache Software License',
                 'Natural Language :: English',
                 'Intended Audience :: Developers',
                 'Topic :: Scientific/Engineering :: Mathematics',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python'],
    **extra
    )



from itertools import chain
from operator import methodcaller


def parse_comma_separated_findall(l):
    '''Returns a flat list of all elements in l that are separated by
       a comma or end-of-list'''
    split = map(methodcaller('split', ','), l)
    flat = chain.from_iterable(split)
    return [s for s in flat if s != '']
